import json
import logging
import time
from json import JSONEncoder

import requests
from requests import exceptions as request_exception

from hackle import version
from hackle.commons import enums


class EventEncoder(JSONEncoder):
    def default(self, o):
        return o.__dict__


class Serializer(object):
    @staticmethod
    def serialize(obj):
        return json.dumps(obj, default=lambda o: o.__dict__.values()[0])


class EventDispatcher(object):
    def __init__(self):
        self.event_url = enums.Default.EVENT_URL

    def dispatch_event(self, sdk_key, event):
        request_headers = {enums.Default.SDK_KEY_HEADER: sdk_key,
                           enums.Default.SDK_NAME_HEADER: 'python-sdk',
                           enums.Default.SDK_VERSION_HEADER: version.__version__,
                           enums.Default.SDK_TIME_HEADER: str(round(time.time() * 1000)),
                           enums.Default.CONTENT_TYPE_HEADER: enums.Default.SDK_CONTENT_TYPE}

        try:
            requests.post(self.event_url, data=json.dumps(event, cls=EventEncoder), headers=request_headers,
                          timeout=enums.Default.REQUEST_TIMEOUT).raise_for_status()

        except request_exception.RequestException as error:
            logging.error('Dispatch event failed. Error: %s' % str(error))



class TargetRuleDeterminer(object):
    def __init__(self, target_matcher):
        self.target_matcher = target_matcher

    def determine_target_rule_or_none(self, workspace, experiment, user):
        for rule in experiment.target_rules:
            if self.target_matcher.matches(rule.target, workspace, user):
                return rule

        return None



import threading
import time
from datetime import timedelta

from six.moves import queue

from hackle import logger as _logging
from hackle.event.event_dispatcher import EventDispatcher
from hackle.event.user_event import (UserEvent, ExposureEvent, TrackEvent)


class BatchEventProcessor(object):
    _DEFAULT_QUEUE_CAPACITY = 1000
    _DEFAULT_BATCH_SIZE = 1000
    _DEFAULT_FLUSH_INTERVAL = 10
    _DEFAULT_TIMEOUT_INTERVAL = 5
    _SHUTDOWN_SIGNAL = object()
    _FLUSH_SIGNAL = object()
    LOCK = threading.Lock()

    def __init__(self, sdk_key=None, event_dispatcher=None, logger=None):
        self.sdk_key = sdk_key
        self.event_dispatcher = event_dispatcher or EventDispatcher
        self.logger = _logging.adapt_logger(logger or _logging.NoOpLogger())
        self.event_queue = queue.Queue(maxsize=self._DEFAULT_QUEUE_CAPACITY)
        self.batch_size = self._DEFAULT_BATCH_SIZE
        self.flush_interval = timedelta(seconds=self._DEFAULT_FLUSH_INTERVAL)
        self.timeout_interval = timedelta(seconds=self._DEFAULT_TIMEOUT_INTERVAL)
        self._current_batch = list()

        self.executor = None
        self.start()

    @property
    def is_running(self):
        return self.executor.is_alive() if self.executor else False

    # noinspection PyMethodMayBeStatic
    def _get_time(self, _time=None):
        if _time is None:
            return time.time()

        return _time

    def start(self):
        if hasattr(self, 'executor') and self.is_running:
            self.logger.warn('BatchEventProcessor already started.')
            return

        self.flushing_interval_deadline = self._get_time() + self._get_time(self.flush_interval.total_seconds())
        self.executor = threading.Thread(target=self._run)
        self.executor.setDaemon(True)
        self.executor.start()

    def _run(self):
        try:
            while True:
                if self._get_time() >= self.flushing_interval_deadline:
                    self._flush_batch()
                    self.flushing_interval_deadline = self._get_time() + \
                                                      self._get_time(self.flush_interval.total_seconds())
                    self.logger.debug('Flush interval deadline. Flushed batch.')

                try:
                    interval = self.flushing_interval_deadline - self._get_time()
                    item = self.event_queue.get(True, interval)

                    if item is None:
                        continue

                except queue.Empty:
                    continue

                if item == self._SHUTDOWN_SIGNAL:
                    self.logger.debug('Shutdown')
                    break

                if item == self._FLUSH_SIGNAL:
                    self.logger.debug('Flush')
                    self._flush_batch()
                    continue

                if isinstance(item, UserEvent):
                    self._add_to_batch(item)

        except Exception as exception:
            self.logger.error('Uncaught exception in event processor: {}'.format(str(exception)))

        finally:
            self.logger.info('Exit Event Processing loop')
            self._flush_batch()

    def flush(self):
        self.event_queue.put(self._FLUSH_SIGNAL)

    def _flush_batch(self):
        batch_len = len(self._current_batch)
        if batch_len == 0:
            self.logger.debug('Noting to flush')
            return

        self.logger.debug('Flushing batch size {}'.format(str(batch_len)))

        with self.LOCK:
            to_process_batch = list(self._current_batch)
            self._current_batch = list()

        exposure_events = []
        track_events = []
        for event in to_process_batch:
            if isinstance(event, ExposureEvent):
                exposure_events.append(event)
            if isinstance(event, TrackEvent):
                track_events.append(event)

        events = {
            'exposureEvents': exposure_events,
            'trackEvents': track_events
        }

        try:
            self.logger.debug('Events : {}'.format(str(events)))
            self.event_dispatcher.dispatch_event(self.sdk_key, events)
        except Exception as exception:
            self.logger.error('Error dispatching event : {} {}'.format(str(events), str(exception)))

    def process(self, event):
        try:
            self.event_queue.put_nowait(event)
        except queue.Full:
            self.logger.debug('Queue is full. Current size : {}'.format(str(self.event_queue.qsize())))

    def _add_to_batch(self, event):
        if len(self._current_batch) == 0:
            self.flushing_interval_deadline = self._get_time() + self._get_time(self.flush_interval.total_seconds())

        with self.LOCK:
            self._current_batch.append(event)
        if len(self._current_batch) >= self.batch_size:
            self.logger.debug('Flushing batch size {}'.format(str(self.batch_size)))
            self._flush_batch()

    def stop(self):
        self.event_queue.put(self._SHUTDOWN_SIGNAL)

        if self.executor:
            self.executor.join(self.timeout_interval.total_seconds())

        if self.is_running:
            self.logger.error('Timeout exceeded {} ms.'.format(str(self.timeout_interval)))



class ContainerResolver(object):

    def __init__(self, bucketer):
        self.bucketer = bucketer

    def is_user_in_container_group(self, container, bucket, experiment, user):
        identifier = user.identifiers.get(experiment.identifier_type)
        if identifier is None:
            return False

        slot = self.bucketer.bucketing(bucket, identifier)
        if slot is None:
            return False

        container_group = container.get_group_or_none(slot.variation_id)
        if container_group is None:
            raise Exception('ContainerGroup[{}]'.format(slot.variation_id))

        return experiment.id in container_group.experiments



import abc

from hackle import logger as _logging
from hackle.evaluation.match.value_operator_matcher import ValueOperatorMatcher, ValueOperatorMatcherFactory


class ConditionMatcher(object, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def matches(self, condition, workspace, user):
        pass


class UserConditionMatcher(ConditionMatcher):
    def __init__(self, user_value_resolver, value_operator_matcher, logger=None):
        self.user_value_resolver = user_value_resolver
        self.value_operator_matcher = value_operator_matcher
        self.logger = _logging.adapt_logger(logger or _logging.NoOpLogger())

    def matches(self, condition, workspace, user):
        user_value = self.user_value_resolver.resolve_or_none(user, condition.key)
        if user_value is None:
            return False

        return self.value_operator_matcher.matches(user_value, condition.match)


class UserValueResolver(object):

    def __init__(self, logger=None):
        self.logger = _logging.adapt_logger(logger or _logging.NoOpLogger())

    def resolve_or_none(self, user, target_key):
        if target_key.type == 'USER_ID':
            return user.identifiers.get(target_key.name)
        elif target_key.type == 'USER_PROPERTY':
            return user.properties.get(target_key.name)
        elif target_key.type == 'HACKLE_PROPERTY':
            return None
        elif target_key.type == 'SEGMENT':
            raise Exception('Unsupported target_key.type [SEGMENT]')
        else:
            self.logger.debug('Unsupported type [{}]. Please use the latest version of sdk.'.format(target_key))
            return None


class SegmentConditionMatcher(ConditionMatcher):

    def __init__(self, segment_matcher, logger=None):
        self.segment_matcher = segment_matcher
        self.logger = _logging.adapt_logger(logger or _logging.NoOpLogger())

    def matches(self, condition, workspace, user):
        if condition.key.type != 'SEGMENT':
            raise Exception('Unsupported target.key.type [{}]'.format(condition.key.type))

        for value in condition.match.values:
            if self._matches(value, workspace, user):
                return self._match_type(condition.match.type, True)

        return self._match_type(condition.match.type, False)

    def _matches(self, value, workspace, user):
        segment_key = value
        if not isinstance(segment_key, str):
            raise Exception('SegmentKey[{}]'.format(segment_key))
        segment = workspace.get_segment_or_none(segment_key)
        if not segment:
            raise Exception('Segment[{}]'.format(segment_key))

        return self.segment_matcher.matches(segment, workspace, user)

    def _match_type(self, match_type, is_matched):
        if match_type == 'MATCH':
            return is_matched
        elif match_type == 'NOT_MATCH':
            return not is_matched
        else:
            self.logger.debug('Unsupported type[{}]. Please use the latest version of sdk.'.format(match_type))
            return False


class SegmentMatcher(object):
    def __init__(self, user_condition_matcher):
        self.user_condition_matcher = user_condition_matcher

    def matches(self, segment, workspace, user):
        for target in segment.targets:
            if self._matches(target, workspace, user):
                return True
        return False

    def _matches(self, target, workspace, user):
        for condition in target.conditions:
            if not self.user_condition_matcher.matches(condition, workspace, user):
                return False
        return True


class ConditionMatcherFactory(object):

    def __init__(self, logger=None):
        self.logger = _logging.adapt_logger(logger or _logging.NoOpLogger())

        self.user_condition_matcher = UserConditionMatcher(
            UserValueResolver(self.logger),
            ValueOperatorMatcher(ValueOperatorMatcherFactory(), self.logger),
            self.logger)
        self.segment_condition_matcher = SegmentConditionMatcher(SegmentMatcher(self.user_condition_matcher),
                                                                 self.logger)

    def get_condition_matcher_or_none(self, target_key_type):
        if target_key_type == 'USER_ID':
            return self.user_condition_matcher
        elif target_key_type == 'USER_PROPERTY':
            return self.user_condition_matcher
        elif target_key_type == 'HACKLE_PROPERTY':
            return self.user_condition_matcher
        elif target_key_type == 'SEGMENT':
            return self.segment_condition_matcher
        else:
            return None



import abc

from hackle.commons import validator
from hackle.evaluation.match.semantic_version import SemanticVersion


class OperatorMatcher(object, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def string_matches(self, user_value, match_value):
        pass

    @abc.abstractmethod
    def number_matches(self, user_value, match_value):
        pass

    @abc.abstractmethod
    def bool_matches(self, user_value, match_value):
        pass

    @abc.abstractmethod
    def version_matches(self, user_value, match_value):
        pass


class InMatcher(OperatorMatcher):
    def string_matches(self, user_value, match_value):
        if isinstance(user_value, str) and isinstance(match_value, str):
            return user_value == match_value
        else:
            return False

    def number_matches(self, user_value, match_value):
        if validator.is_finite_number(user_value) and validator.is_finite_number(match_value):
            return user_value == match_value
        else:
            return False

    def bool_matches(self, user_value, match_value):
        if isinstance(user_value, bool) and isinstance(match_value, bool):
            return user_value == match_value
        else:
            return False

    def version_matches(self, user_value, match_value):
        if isinstance(user_value, SemanticVersion) and isinstance(match_value, SemanticVersion):
            return user_value == match_value
        else:
            return False


class ContainsMatcher(OperatorMatcher):
    def string_matches(self, user_value, match_value):
        if isinstance(user_value, str) and isinstance(match_value, str):
            return match_value in user_value
        else:
            return False

    def number_matches(self, user_value, match_value):
        return False

    def bool_matches(self, user_value, match_value):
        return False

    def version_matches(self, user_value, match_value):
        return False


class StartsWithMatcher(OperatorMatcher):
    def string_matches(self, user_value, match_value):
        if isinstance(user_value, str) and isinstance(match_value, str):
            return user_value.startswith(match_value)
        else:
            return False

    def number_matches(self, user_value, match_value):
        return False

    def bool_matches(self, user_value, match_value):
        return False

    def version_matches(self, user_value, match_value):
        return False


class EndsWithMatcher(OperatorMatcher):
    def string_matches(self, user_value, match_value):
        if isinstance(user_value, str) and isinstance(match_value, str):
            return user_value.endswith(match_value)
        else:
            return False

    def number_matches(self, user_value, match_value):
        return False

    def bool_matches(self, user_value, match_value):
        return False

    def version_matches(self, user_value, match_value):
        return False


class GreaterThanMatcher(OperatorMatcher):
    def string_matches(self, user_value, match_value):
        return False

    def number_matches(self, user_value, match_value):
        if validator.is_finite_number(user_value) and validator.is_finite_number(match_value):
            return user_value > match_value
        else:
            return False

    def bool_matches(self, user_value, match_value):
        return False

    def version_matches(self, user_value, match_value):
        if isinstance(user_value, SemanticVersion) and isinstance(match_value, SemanticVersion):
            return user_value > match_value
        else:
            return False


class GreaterThanOrEqualToMatcher(OperatorMatcher):
    def string_matches(self, user_value, match_value):
        return False

    def number_matches(self, user_value, match_value):
        if validator.is_finite_number(user_value) and validator.is_finite_number(match_value):
            return user_value >= match_value
        else:
            return False

    def bool_matches(self, user_value, match_value):
        return False

    def version_matches(self, user_value, match_value):
        if isinstance(user_value, SemanticVersion) and isinstance(match_value, SemanticVersion):
            return user_value >= match_value
        else:
            return False


class LessThanMatcher(OperatorMatcher):
    def string_matches(self, user_value, match_value):
        return False

    def number_matches(self, user_value, match_value):
        if validator.is_finite_number(user_value) and validator.is_finite_number(match_value):
            return user_value < match_value
        else:
            return False

    def bool_matches(self, user_value, match_value):
        return False

    def version_matches(self, user_value, match_value):
        if isinstance(user_value, SemanticVersion) and isinstance(match_value, SemanticVersion):
            return user_value < match_value
        else:
            return False


class LessThanOrEqualToMatcher(OperatorMatcher):
    def string_matches(self, user_value, match_value):
        return False

    def number_matches(self, user_value, match_value):
        if validator.is_finite_number(user_value) and validator.is_finite_number(match_value):
            return user_value <= match_value
        else:
            return False

    def bool_matches(self, user_value, match_value):
        return False

    def version_matches(self, user_value, match_value):
        if isinstance(user_value, SemanticVersion) and isinstance(match_value, SemanticVersion):
            return user_value <= match_value
        else:
            return False



from hackle import exceptions as hackle_exceptions
from hackle import logger as _logging
from hackle.commons import utils
from hackle.commons import validator
from hackle.decision import ExperimentDecision, DecisionReason, FeatureFlagDecision
from hackle.evaluation.evaluator import Evaluator
from hackle.evaluation.flow.evaluation_flow_factory import EvaluationFlowFactory
from hackle.event.event_dispatcher import EventDispatcher
from hackle.event.event_processor import BatchEventProcessor
from hackle.internal_client import InternalClient
from hackle.user.hackle_user_resolver import HackleUserResolver
from hackle.workspace_fetcher import WorkspaceFetcher


def __singleton(cls):
    instance = [None]

    def wrapper(*args, **kwargs):
        if instance[0] is None:
            instance[0] = cls(*args, **kwargs)
        return instance[0]

    return wrapper


@__singleton
class Client(object):
    def __init__(self, sdk_key=None, logger=None, timeout=None):
        if sdk_key is None:
            raise hackle_exceptions.RequiredParameterException('sdk_key must not be empty.')

        self.logger = _logging.adapt_logger(logger or _logging.NoOpLogger())

        self.internal_client = InternalClient(
            evaluator=Evaluator(EvaluationFlowFactory(self.logger)),
            workspace_fetcher=WorkspaceFetcher(sdk_key, logger=self.logger, timeout=timeout),
            event_processor=BatchEventProcessor(sdk_key, EventDispatcher(), self.logger),
            logger=self.logger
        )

        self.hackle_user_resolver = HackleUserResolver(logger)

    def close(self):
        self.internal_client.close()

    def __exit__(self):
        self.close()

    def variation(self, experiment_key, user, default_variation='A'):
        """
        Decide the variation to expose to the user for experiment.

        This method return the "A" if:
            - The experiment key is invalid
            - The experiment has not started yet
            - The user is not allocated to the experiment
            - The decided variation has been dropped

        :param int experiment_key: the unique key of the experiment.
        :param hackle.model.User or hackle.model.HackleUser user: the user to participate in the experiment.
        :param str default_variation: the default variation of the experiment.

        :return: the decided variation for the user, or the default variation.
        """
        return self.variation_detail(experiment_key, user, default_variation).variation

    def variation_detail(self, experiment_key, user, default_variation='A'):
        """
        Decide the variation to expose to the user for experiment, and returns an object that
        describes the way the variation was decided.

        :param int experiment_key: the unique key of the experiment.
        :param hackle.model.User or hackle.model.HackleUser user: the user to participate in the experiment.
        :param str default_variation: the default variation of the experiment.

        :return: a object describing the result
        """
        try:
            if not validator.is_non_zero_and_empty_int(experiment_key):
                self.logger.error('Experiment Key must not be empty. : {}'.format(experiment_key))
                return ExperimentDecision(default_variation, DecisionReason.INVALID_INPUT)

            hackle_user = self.hackle_user_resolver.resolve_or_none(user)
            if hackle_user is None:
                return ExperimentDecision(default_variation, DecisionReason.INVALID_INPUT)

            return self.internal_client.experiment(experiment_key, hackle_user, default_variation)
        except Exception as e:
            self.logger.error(
                "Unexpected error while deciding variation of experiment[{}]: {}".format(experiment_key, str(e)))
            return ExperimentDecision(default_variation, DecisionReason.EXCEPTION)

    def is_feature_on(self, feature_key, user):
        """
        Decide whether the feature is turned on to the user.

        :param int feature_key: the unique key of the feature.
        :param hackle.model.User or hackle.model.HackleUser user: the user to participate in the experiment.
        :return: True if the feature is on
                  False if the feature is off
        """
        return self.feature_flag_detail(feature_key, user).is_on

    def feature_flag_detail(self, feature_key, user):
        """
        Decide whether the feature is turned on to the user, and returns an object that
        describes the way the value was decided.

        :param int feature_key: the unique key of the feature.
        :param hackle.model.User or hackle.model.HackleUser user: the user to participate in the experiment.

        :return: a object describing the result
        """
        try:

            if not validator.is_non_zero_and_empty_int(feature_key):
                self.logger.error('Experiment Key must not be empty. : {}'.format(feature_key))
                return FeatureFlagDecision(False, DecisionReason.INVALID_INPUT)

            hackle_user = self.hackle_user_resolver.resolve_or_none(user)
            if hackle_user is None:
                return FeatureFlagDecision(False, DecisionReason.INVALID_INPUT)

            return self.internal_client.feature_flag(feature_key, hackle_user)
        except Exception as e:
            self.logger.error("Unexpected error while deciding feature flag[{}]: {}".format(feature_key, str(e)))
            return FeatureFlagDecision(False, DecisionReason.EXCEPTION)

    def track(self, event, user):
        """
        Records the event performed by the user with additional numeric value.

        :param hackle.model.Event event: the unique key of the event. MUST NOT be null.
        :param hackle.model.User or hackle.model.HackleUser user: the user to participate in the experiment.
        """
        try:
            if not validator.is_valid_event(event):
                self.logger.error('Event is not valid. Event must be hackle.model.event and event.id\'s type must be '
                                  'string_types. value\'s type must be numeric. '
                                  ': {}'.format(event))
                return

            if not validator.is_valid_properties(event.properties):
                self.logger.warning(
                    'Event properties is not valid. Event properties must be not dic types and items types must be string_types, number, bool.'
                    ': {}'.format(event.properties))
                event.properties = utils.filter_properties(event.properties)

            hackle_user = self.hackle_user_resolver.resolve_or_none(user)
            if hackle_user is None:
                return

            self.internal_client.track_event(event, hackle_user)
        except Exception as e:
            self.logger.error('Unexpected error while tracking event: {}'.format(str(e)))



from hackle import logger as _logging


class ActionResolver(object):
    def __init__(self, bucketer, logger=None):
        self.bucketer = bucketer
        self.logger = _logging.adapt_logger(logger or _logging.NoOpLogger())

    def resolve_or_none(self, action, workspace, experiment, user):
        if action.type == 'VARIATION':
            return self._resolve_variation(action, experiment)
        elif action.type == 'BUCKET':
            return self._resolve_bucket(action, workspace, experiment, user)
        else:
            self.logger.debug('Unsupported type[{}]. Please use the latest version of sdk.'.format(action.type))
            return None

    def _resolve_variation(self, action, experiment):
        variation = experiment.get_variation_by_id_or_none(action.variation_id)
        if variation is None:
            raise Exception('variation[{}]'.format(action.variation_id))
        return variation

    def _resolve_bucket(self, action, workspace, experiment, user):
        bucket = workspace.get_bucket_or_none(action.bucket_id)
        if bucket is None:
            raise Exception('bucket[{}]'.format(action.bucket_id))

        identifier = user.identifiers.get(experiment.identifier_type)
        if identifier is None:
            return None

        allocated_slot = self.bucketer.bucketing(bucket, identifier)
        if allocated_slot is None:
            return None

        return experiment.get_variation_by_id_or_none(allocated_slot.variation_id)



from hackle import logger as _logging


class TargetMatcher(object):

    def __init__(self, condition_matcher_factory, logger=None):
        self.condition_matcher_factory = condition_matcher_factory
        self.logger = _logging.adapt_logger(logger or _logging.NoOpLogger())

    def matches(self, target, workspace, user):
        for condition in target.conditions:
            if not self._matches(condition, workspace, user):
                return False
        return True

    def _matches(self, condition, workspace, user):
        condition_matcher = self.condition_matcher_factory.get_condition_matcher_or_none(condition.key.type)
        if not condition_matcher:
            self.logger.debug('Unsupported type [{}]. Please use the latest version of sdk.'.format(condition.key.type))
            return False

        return condition_matcher.matches(condition, workspace, user)



class ExperimentTargetDeterminer(object):
    def __init__(self, target_matcher):
        self.target_matcher = target_matcher

    def is_user_in_experiment_target(self, workspace, experiment, user):
        if not experiment.target_audiences:
            return True

        for target in experiment.target_audiences:
            if self.target_matcher.matches(target, workspace, user):
                return True

        return False



import math
import numbers

from six import integer_types
from six import string_types

from hackle.model import Event


def is_non_empty_string(input_id_key):
    if input_id_key and isinstance(input_id_key, string_types):
        return True

    return False


def is_non_zero_and_empty_int(input_id_key):
    if input_id_key and isinstance(input_id_key, integer_types) and input_id_key > 0:
        return True

    return False


def is_number(input_id_key):
    if input_id_key and isinstance(input_id_key, numbers.Number):
        return True

    return False


def is_event_value_valid(attribute_value):
    if attribute_value is None:
        return True

    if isinstance(attribute_value, (numbers.Integral, float)):
        return is_finite_number(attribute_value)

    return False


def is_finite_number(value):
    if not isinstance(value, (numbers.Integral, float)):
        # numbers.Integral instead of int to accommodate long integer in python 2
        return False

    if isinstance(value, bool):
        # bool is a subclass of int
        return False

    if isinstance(value, float):
        if math.isnan(value) or math.isinf(value):
            return False

    if abs(value) > (2 ** 53):
        return False

    return True


def is_valid_event(event):
    if event is None:
        return False

    if event and not isinstance(event, Event):
        return False

    if event and event.key is None:
        return False

    if event and not isinstance(event.key, string_types):
        return False

    if event and event.value and not is_event_value_valid(event.value):
        return False

    return True


def is_valid_properties(properties):
    if properties and not isinstance(properties, dict):
        return False

    for property_key in properties.keys():
        property_value = properties.get(property_key)
        if not is_property_valid(property_key, property_value):
            return False

    return True


def is_property_valid(property_key, property_value):
    if not isinstance(property_key, string_types):
        return False

    if len(property_key) > 128:
        return False

    if isinstance(property_value, (string_types, bool)):
        return True

    if isinstance(property_value, (numbers.Integral, float)):
        return is_finite_number(property_value)

    return False



try:
    import mmh3
except ImportError:
    from lib import pymmh3 as mmh3


class Bucketer(object):
    def __init__(self):
        return

    def bucketing(self, bucket, identifier):
        slot_number = self._calculate_slot_number(identifier, bucket.seed, bucket.slot_size)
        return bucket.get_slot_or_none(slot_number)

    def _calculate_slot_number(self, identifier, seed, slot_size):
        hash_value = mmh3.hash(identifier, seed)
        return abs(hash_value) % abs(slot_size)



class OverrideResolver(object):
    def __init__(self, target_matcher, action_resolver):
        self.target_matcher = target_matcher
        self.action_resolver = action_resolver

    def resolve_or_none(self, workspace, experiment, user):
        user_overridden_variation = self._resolve_user_override_or_none(experiment, user)
        if user_overridden_variation is not None:
            return user_overridden_variation

        return self._resolve_segment_override(workspace, experiment, user)

    def _resolve_user_override_or_none(self, experiment, user):
        identifier = user.identifiers.get(experiment.identifier_type)
        if identifier is None:
            return None

        overridden_variation_id = experiment.user_overrides.get(identifier)
        if overridden_variation_id is None:
            return None

        return experiment.get_variation_by_id_or_none(overridden_variation_id)

    def _resolve_segment_override(self, workspace, experiment, user):
        for overridden_rule in experiment.segment_overrides:
            if self.target_matcher.matches(overridden_rule.target, workspace, user):
                return self.action_resolver.resolve_or_none(overridden_rule.action, workspace, experiment, user)

        return None



import abc

from hackle.commons import validator
from hackle.evaluation.match.semantic_version import SemanticVersion


class ValueMatcher(object, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def matches(self, operator_matcher, user_value, match_value):
        pass


class StringValueMatcher(ValueMatcher):
    def matches(self, operator_matcher, user_value, match_value):
        if isinstance(user_value, str) and isinstance(match_value, str):
            return operator_matcher.string_matches(user_value, match_value)
        else:
            return False


class NumberValueMatcher(ValueMatcher):
    def matches(self, operator_matcher, user_value, match_value):
        if validator.is_finite_number(user_value) and validator.is_finite_number(match_value):
            return operator_matcher.number_matches(user_value, match_value)
        else:
            return False


class BoolValueMatcher(ValueMatcher):
    def matches(self, operator_matcher, user_value, match_value):
        if isinstance(user_value, bool) and isinstance(match_value, bool):
            return operator_matcher.bool_matches(user_value, match_value)
        else:
            return False


class VersionValueMatcher(ValueMatcher):
    def matches(self, operator_matcher, user_value, match_value):
        user_version = SemanticVersion.parse_or_none(user_value)
        match_version = SemanticVersion.parse_or_none(match_value)
        if user_version is not None and match_version is not None:
            return operator_matcher.version_matches(user_version, match_version)
        else:
            return False



import re

from six import string_types


def compare_values(a, b):
    return (a > b) - (a < b)


def compare_identifiers(identifier1, identifier2):
    if identifier1.isnumeric() and identifier2.isnumeric():
        return compare_values(int(identifier1), int(identifier2))
    else:
        return compare_values(identifier1, identifier2)


class SemanticVersion(object):
    _REGEX = re.compile(
        r"""
            ^(0|[1-9]\d*)
            (?:\.(0|[1-9]\d*))?
            (?:\.(0|[1-9]\d*))?
            (?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?
            (?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?
            $
        """,
        re.VERBOSE,
    )

    def __init__(self, core_version, prerelease, build):
        self.core_version = core_version
        self.prerelease = prerelease
        self.build = build

    def compare(self, other):
        result = self.core_version.compare(other.core_version)
        if result != 0:
            return result
        else:
            return self.prerelease.compare(other.prerelease)

    def __str__(self):
        version = str(self.core_version)
        if self.prerelease.is_not_empty():
            version += "-{}".format(self.prerelease)
        if self.build.is_not_empty():
            version += "+{}".format(self.build)
        return version

    def __eq__(self, other):
        return self.compare(other) == 0

    def __ne__(self, other):
        return self.compare(other) != 0

    def __gt__(self, other):
        return self.compare(other) > 0

    def __ge__(self, other):
        return self.compare(other) >= 0

    def __le__(self, other):
        return self.compare(other) <= 0

    def __lt__(self, other):
        return self.compare(other) < 0

    @staticmethod
    def parse_or_none(value):
        if not isinstance(value, string_types):
            return None

        match = SemanticVersion._REGEX.match(value)
        if match is None:
            return None

        groups = match.groups()

        major = groups[0]
        minor = groups[1]
        patch = groups[2]

        core_version = CoreVersion(
            major=int(major),
            minor=int(minor) if minor is not None else 0,
            patch=int(patch) if patch is not None else 0,
        )
        prerelease = MetadataVersion.parse(groups[3])
        build = MetadataVersion.parse(groups[4])

        return SemanticVersion(core_version, prerelease, build)


class CoreVersion(object):

    def __init__(self, major, minor, patch):
        self.major = major
        self.minor = minor
        self.patch = patch

    def compare(self, other):
        a = (self.major, self.minor, self.patch)
        b = (other.major, other.minor, other.patch)
        return compare_values(a, b)

    def __str__(self):
        return "{}.{}.{}".format(self.major, self.minor, self.patch)


class MetadataVersion(object):

    def __init__(self, identifiers):
        self.identifiers = identifiers

    def is_empty(self):
        return len(self.identifiers) == 0

    def is_not_empty(self):
        return not self.is_empty()

    def compare(self, other):
        if self.is_empty() and other.is_empty():
            return 0

        if self.is_empty() and other.is_not_empty():
            return 1

        if self.is_not_empty() and other.is_empty():
            return -1

        return self.compare_identifiers(other)

    def compare_identifiers(self, other):
        for self_identifier, other_identifier in zip(self.identifiers, other.identifiers):
            result = compare_identifiers(self_identifier, other_identifier)
            if result != 0:
                return result
        else:
            return compare_values(len(self.identifiers), len(other.identifiers))

    def __str__(self):
        return ".".join(self.identifiers)

    @staticmethod
    def parse(value):
        if value is None:
            return MetadataVersion([])
        else:
            return MetadataVersion(value.split("."))



from hackle.decision import DecisionReason
from hackle.evaluation.evaluator import Evaluation


class EvaluationFlow(object):

    def __init__(self, flow_evaluator=None, next_flow=None):
        self.flow_evaluator = flow_evaluator
        self.next_flow = next_flow

    def evaluate(self, workspace, experiment, user, default_variation_key):
        if self.is_end():
            return Evaluation.of(experiment, default_variation_key, DecisionReason.TRAFFIC_NOT_ALLOCATED)
        else:
            return self.flow_evaluator.evaluate(workspace, experiment, user, default_variation_key, self.next_flow)

    def is_end(self):
        return self.flow_evaluator is None or self.next_flow is None

    @staticmethod
    def of(*flow_evaluators):
        flow = EvaluationFlow()
        for flow_evaluator in reversed(flow_evaluators):
            flow = EvaluationFlow(flow_evaluator, flow)
        return flow



from hackle import logger as _logging


class IdentifiersBuilder(object):

    def __init__(self, logger=None):
        self.logger = _logging.adapt_logger(logger or _logging.NoOpLogger())
        self.identifiers = {}

    def add_identifiers(self, identifiers):
        if identifiers is None:
            return self

        if not isinstance(identifiers, dict):
            self.logger.warning('Identifiers must be dictionary.')
            return self

        for identifier_type in identifiers:
            self.add(identifier_type, identifiers[identifier_type])

        return self

    def add(self, identifier_type, identifier_value):
        if self._is_valid(identifier_type, identifier_value):
            self.identifiers[identifier_type] = identifier_value
        else:
            self.logger.warning('Invalid user identifier [type={}, value={}]'.format(identifier_type, identifier_value))
        return self

    def _is_valid(self, identifier_type, identifier_value):
        if identifier_type is None:
            return False

        if not isinstance(identifier_type, str):
            return False

        if len(identifier_type) == 0:
            return False

        if len(identifier_type) > 128:
            return False

        if identifier_value is None:
            return False

        if not isinstance(identifier_value, str):
            return False

        if len(identifier_value) == 0:
            return False

        if len(identifier_value) > 512:
            return False

        return True

    def build(self):
        return self.identifiers



import abc

from hackle.decision import DecisionReason
from hackle.evaluation.evaluator import Evaluation


class FlowEvaluator(object, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def evaluate(self, workspace, experiment, user, default_variation_key, next_flow):
        pass


class OverrideEvaluator(FlowEvaluator):

    def __init__(self, override_resolver):
        self.override_resolver = override_resolver

    def evaluate(self, workspace, experiment, user, default_variation_key, next_flow):
        overridden_variation = self.override_resolver.resolve_or_none(workspace, experiment, user)

        if overridden_variation:
            if experiment.type == 'AB_TEST':
                return Evaluation.with_variation(overridden_variation, DecisionReason.OVERRIDDEN)
            elif experiment.type == 'FEATURE_FLAG':
                return Evaluation.with_variation(overridden_variation, DecisionReason.INDIVIDUAL_TARGET_MATCH)
            else:
                raise Exception('experiment type [{}]'.format(experiment.type))
        else:
            return next_flow.evaluate(workspace, experiment, user, default_variation_key)


class DraftEvaluator(FlowEvaluator):
    def evaluate(self, workspace, experiment, user, default_variation_key, next_flow):
        if experiment.status == 'DRAFT':
            return Evaluation.of(experiment, default_variation_key, DecisionReason.EXPERIMENT_DRAFT)
        else:
            return next_flow.evaluate(workspace, experiment, user, default_variation_key)


class PausedEvaluator(FlowEvaluator):
    def evaluate(self, workspace, experiment, user, default_variation_key, next_flow):
        if experiment.status == 'PAUSED':
            if experiment.type == 'AB_TEST':
                return Evaluation.of(experiment, default_variation_key, DecisionReason.EXPERIMENT_PAUSED)
            elif experiment.type == 'FEATURE_FLAG':
                return Evaluation.of(experiment, default_variation_key, DecisionReason.FEATURE_FLAG_INACTIVE)
            else:
                raise Exception('experiment type [{}]'.format(experiment.type))
        else:
            return next_flow.evaluate(workspace, experiment, user, default_variation_key)


class CompletedEvaluator(FlowEvaluator):
    def evaluate(self, workspace, experiment, user, default_variation_key, next_flow):
        if experiment.status == 'COMPLETED':
            return Evaluation.with_variation(experiment.winner_variation, DecisionReason.EXPERIMENT_COMPLETED)
        else:
            return next_flow.evaluate(workspace, experiment, user, default_variation_key)


class ExperimentTargetEvaluator(FlowEvaluator):

    def __init__(self, experiment_target_determiner):
        self.experiment_target_determiner = experiment_target_determiner

    def evaluate(self, workspace, experiment, user, default_variation_key, next_flow):
        if experiment.type != 'AB_TEST':
            raise Exception('experiment type must be AB_TEST [{}]'.format(experiment.id))

        is_user_in_experiment_target = self.experiment_target_determiner.is_user_in_experiment_target(workspace,
                                                                                                      experiment, user)
        if is_user_in_experiment_target:
            return next_flow.evaluate(workspace, experiment, user, default_variation_key)
        else:
            return Evaluation.of(experiment, default_variation_key, DecisionReason.NOT_IN_EXPERIMENT_TARGET)


class TrafficAllocateEvaluator(FlowEvaluator):

    def __init__(self, action_resolver):
        self.action_resolver = action_resolver

    def evaluate(self, workspace, experiment, user, default_variation_key, next_flow):
        if experiment.status != 'RUNNING':
            raise Exception('experiment status must be RUNNING [{}]'.format(experiment.id))

        if experiment.type != 'AB_TEST':
            raise Exception('experiment type must be AB_TEST [{}]'.format(experiment.id))

        variation = self.action_resolver.resolve_or_none(experiment.default_rule, workspace, experiment, user)
        if not variation:
            return Evaluation.of(experiment, default_variation_key, DecisionReason.TRAFFIC_NOT_ALLOCATED)

        if variation.is_dropped:
            return Evaluation.of(experiment, default_variation_key, DecisionReason.VARIATION_DROPPED)

        return Evaluation.with_variation(variation, DecisionReason.TRAFFIC_ALLOCATED)


class TargetRuleEvaluator(FlowEvaluator):

    def __init__(self, target_rule_determiner, action_resolver):
        self.target_rule_determiner = target_rule_determiner
        self.action_resolver = action_resolver

    def evaluate(self, workspace, experiment, user, default_variation_key, next_flow):
        if experiment.status != 'RUNNING':
            raise Exception('experiment status must be RUNNING [{}]'.format(experiment.id))

        if experiment.type != 'FEATURE_FLAG':
            raise Exception('experiment type must be FEATURE_FLAG [{}]'.format(experiment.id))

        if user.identifiers.get(experiment.identifier_type) is None:
            return next_flow.evaluate(workspace, experiment, user, default_variation_key)

        target_rule = self.target_rule_determiner.determine_target_rule_or_none(workspace, experiment, user)
        if target_rule is None:
            return next_flow.evaluate(workspace, experiment, user, default_variation_key)

        variation = self.action_resolver.resolve_or_none(target_rule.action, workspace, experiment, user)
        if not variation:
            raise Exception('FeatureFlag must decide the variation [{}]'.format(experiment.id))

        return Evaluation.with_variation(variation, DecisionReason.TARGET_RULE_MATCH)


class DefaultRuleEvaluator(FlowEvaluator):

    def __init__(self, action_resolver):
        self.action_resolver = action_resolver

    def evaluate(self, workspace, experiment, user, default_variation_key, next_flow):
        if experiment.status != 'RUNNING':
            raise Exception('experiment status must be RUNNING [{}]'.format(experiment.id))

        if experiment.type != 'FEATURE_FLAG':
            raise Exception('experiment type must be FEATURE_FLAG [{}]'.format(experiment.id))

        if user.identifiers.get(experiment.identifier_type) is None:
            return Evaluation.of(experiment, default_variation_key, DecisionReason.DEFAULT_RULE)

        variation = self.action_resolver.resolve_or_none(experiment.default_rule, workspace, experiment, user)
        if not variation:
            raise Exception('FeatureFlag must decide the variation [{}]'.format(experiment.id))

        return Evaluation.with_variation(variation, DecisionReason.DEFAULT_RULE)


class ContainerEvaluator(FlowEvaluator):

    def __init__(self, container_resolver):
        self.container_resolver = container_resolver

    def evaluate(self, workspace, experiment, user, default_variation_key, next_flow):
        container_id = experiment.container_id
        if container_id is None:
            return next_flow.evaluate(workspace, experiment, user, default_variation_key)

        container = workspace.get_container_or_none(container_id)
        if container is None:
            raise Exception('Container[{}]'.format(container_id))

        bucket = workspace.get_bucket_or_none(container.bucket_id)
        if bucket is None:
            raise Exception('Bucket[{}]'.format(container.bucket_id))

        is_user_in_container_group = self.container_resolver.is_user_in_container_group(container, bucket, experiment,
                                                                                        user)
        if is_user_in_container_group:
            return next_flow.evaluate(workspace, experiment, user, default_variation_key)
        else:
            return Evaluation.of(experiment, default_variation_key, DecisionReason.NOT_IN_MUTUAL_EXCLUSION_EXPERIMENT)


class IdentifierEvaluator(FlowEvaluator):
    def evaluate(self, workspace, experiment, user, default_variation_key, next_flow):
        if user.identifiers.get(experiment.identifier_type) is not None:
            return next_flow.evaluate(workspace, experiment, user, default_variation_key)
        else:
            return Evaluation.of(experiment, default_variation_key, DecisionReason.IDENTIFIER_NOT_FOUND)



class ExperimentDecision:
    def __init__(self, variation, reason):
        self.variation = variation
        self.reason = reason

    def __eq__(self, o):
        return self.variation == o.variation and self.reason == o.reason

    def __ne__(self, o):
        return not self.__eq__(o)

    def __str__(self):
        return '(variation={}, reason={})'.format(self.variation, self.reason)

    def __repr__(self):
        return self.__str__()


class FeatureFlagDecision:
    def __init__(self, is_on, reason):
        self.is_on = is_on
        self.reason = reason

    def __eq__(self, o):
        return self.is_on == o.is_on and self.reason == o.reason

    def __ne__(self, o):
        return not self.__eq__(o)

    def __str__(self):
        return '(is_on={}, reason={})'.format(self.is_on, self.reason)

    def __repr__(self):
        return self.__str__()


class DecisionReason(object):
    SDK_NOK_READY = 'SDK_NOK_READY'
    EXCEPTION = 'EXCEPTION'
    INVALID_INPUT = 'INVALID_INPUT'

    EXPERIMENT_NOT_FOUND = 'EXPERIMENT_NOT_FOUND'
    EXPERIMENT_DRAFT = 'EXPERIMENT_DRAFT'
    EXPERIMENT_PAUSED = 'EXPERIMENT_PAUSED'
    EXPERIMENT_COMPLETED = 'EXPERIMENT_COMPLETED'
    OVERRIDDEN = 'OVERRIDDEN'
    TRAFFIC_NOT_ALLOCATED = 'TRAFFIC_NOT_ALLOCATED'
    TRAFFIC_ALLOCATED = 'TRAFFIC_ALLOCATED'
    VARIATION_DROPPED = 'VARIATION_DROPPED'
    NOT_IN_EXPERIMENT_TARGET = 'NOT_IN_EXPERIMENT_TARGET'
    NOT_IN_MUTUAL_EXCLUSION_EXPERIMENT = 'NOT_IN_MUTUAL_EXCLUSION_EXPERIMENT'
    IDENTIFIER_NOT_FOUND = 'IDENTIFIER_NOT_FOUND'

    FEATURE_FLAG_NOT_FOUND = 'FEATURE_FLAG_NOT_FOUND'
    FEATURE_FLAG_INACTIVE = 'FEATURE_FLAG_INACTIVE'
    INDIVIDUAL_TARGET_MATCH = 'INDIVIDUAL_TARGET_MATCH'
    TARGET_RULE_MATCH = 'TARGET_RULE_MATCH'
    DEFAULT_RULE = 'DEFAULT_RULE'



class InternalHackleUser(object):
    def __init__(self, identifiers, properties):
        self.identifiers = identifiers
        self.properties = properties

    def __eq__(self, other):
        if not isinstance(other, InternalHackleUser):
            return False
        return self.identifiers == other.identifiers and self.properties == other.properties

    def __ne__(self, other):
        return not self.__eq__(other)



class User(object):
    def __init__(self, id, properties):
        self.id = id
        self.properties = properties

    def __str__(self):
        return 'User(id={}, properties={})'.format(self.id, self.properties)


class HackleUser(object):
    def __init__(self, id=None, user_id=None, device_id=None, identifiers=None, properties=None):
        self.id = id
        self.user_id = user_id
        self.device_id = device_id
        self.identifiers = identifiers
        self.properties = properties

    def __str__(self):
        return 'HackleUser(id={}, user_id={}, device_id={}, identifiers={}, properties={})'.format(self.id,
                                                                                                   self.user_id,
                                                                                                   self.device_id,
                                                                                                   self.identifiers,
                                                                                                   self.properties)

    @staticmethod
    def of(user):
        return HackleUser(user.id, None, None, None, user.properties)


class Event(object):
    def __init__(self, key, value, properties):
        self.key = key
        self.value = value
        self.properties = properties


class Hackle:
    @staticmethod
    def user(id, **kwargs):
        return User(id, kwargs)

    @staticmethod
    def event(key, value=None, **kwargs):
        return Event(key, value, kwargs)



from hackle import logger as _logging
from hackle.evaluation.action.action_resolver import ActionResolver
from hackle.evaluation.bucket.bucketer import Bucketer
from hackle.evaluation.container.container_resolver import ContainerResolver
from hackle.evaluation.flow.evaluation_flow import EvaluationFlow
from hackle.evaluation.flow.flow_evaluator import *
from hackle.evaluation.match.condition_matcher import ConditionMatcherFactory
from hackle.evaluation.match.target_matcher import TargetMatcher
from hackle.evaluation.target.experiment_target_determiner import ExperimentTargetDeterminer
from hackle.evaluation.target.override_resolver import OverrideResolver
from hackle.evaluation.target.target_rule_determiner import TargetRuleDeterminer


class EvaluationFlowFactory(object):

    def __init__(self, logger=None):
        self.logger = _logging.adapt_logger(logger or _logging.NoOpLogger())

        bucketer = Bucketer()
        target_matcher = TargetMatcher(ConditionMatcherFactory())
        action_resolver = ActionResolver(bucketer, self.logger)
        override_resolver = OverrideResolver(target_matcher, action_resolver)
        container_resolver = ContainerResolver(bucketer)

        self.ab_test_flow = EvaluationFlow.of(
            OverrideEvaluator(override_resolver),
            IdentifierEvaluator(),
            ContainerEvaluator(container_resolver),
            ExperimentTargetEvaluator(ExperimentTargetDeterminer(target_matcher)),
            DraftEvaluator(),
            PausedEvaluator(),
            CompletedEvaluator(),
            TrafficAllocateEvaluator(action_resolver)
        )

        self.feature_flag_flow = EvaluationFlow.of(
            DraftEvaluator(),
            PausedEvaluator(),
            CompletedEvaluator(),
            OverrideEvaluator(override_resolver),
            IdentifierEvaluator(),
            TargetRuleEvaluator(TargetRuleDeterminer(target_matcher), action_resolver),
            DefaultRuleEvaluator(action_resolver)
        )

    def get_evaluation_flow(self, experiment_type):
        if experiment_type == 'AB_TEST':
            return self.ab_test_flow
        elif experiment_type == 'FEATURE_FLAG':
            return self.feature_flag_flow
        else:
            raise Exception('Unsupported type [{}]'.format(experiment_type))



from hackle.decision import ExperimentDecision, FeatureFlagDecision, DecisionReason
from hackle.entities import EventType
from hackle.event.user_event import UserEvent


class InternalClient(object):

    def __init__(self, evaluator, workspace_fetcher, event_processor, logger):
        self.evaluator = evaluator
        self.workspace_fetcher = workspace_fetcher
        self.event_processor = event_processor
        self.logger = logger

    def close(self):
        self.workspace_fetcher.stop()
        self.event_processor.stop()

    def experiment(self, experiment_key, user, default_variation_key):
        workspace = self.workspace_fetcher.get_workspace()
        if workspace is None:
            return ExperimentDecision(default_variation_key, DecisionReason.SDK_NOK_READY)

        experiment = workspace.get_experiment_or_none(experiment_key)
        if experiment is None:
            return ExperimentDecision(default_variation_key, DecisionReason.EXPERIMENT_NOT_FOUND)

        evaluation = self.evaluator.evaluate(workspace, experiment, user, default_variation_key)
        self.event_processor.process(UserEvent.exposure(experiment, user, evaluation))

        return ExperimentDecision(evaluation.variation_key, evaluation.reason)

    def feature_flag(self, feature_key, user):
        workspace = self.workspace_fetcher.get_workspace()
        if workspace is None:
            return FeatureFlagDecision(False, DecisionReason.SDK_NOK_READY)

        feature_flag = workspace.get_feature_flag_or_none(feature_key)
        if feature_flag is None:
            return FeatureFlagDecision(False, DecisionReason.FEATURE_FLAG_NOT_FOUND)

        evaluation = self.evaluator.evaluate(workspace, feature_flag, user, 'A')
        self.event_processor.process(UserEvent.exposure(feature_flag, user, evaluation))

        if evaluation.variation_key == 'A':
            return FeatureFlagDecision(False, evaluation.reason)
        else:
            return FeatureFlagDecision(True, evaluation.reason)

    def track_event(self, event, user):
        event_type = self._event_type(event)
        self.event_processor.process(UserEvent.track(event_type, event, user))
        return

    def _event_type(self, event):
        config = self.workspace_fetcher.get_workspace()

        if not config:
            return EventType(0, event.key)

        event_type = config.get_event_type_or_none(event.key)

        if not event_type:
            return EventType(0, event.key)

        return event_type



from hackle import logger as _logging
from hackle.commons import utils
from hackle.commons import validator
from hackle.model import User, HackleUser
from hackle.user.identifier_type import IdentifierType
from hackle.user.identifiers_builder import IdentifiersBuilder
from hackle.user.internal_hackle_user import InternalHackleUser


class HackleUserResolver(object):

    def __init__(self, logger=None):
        self.logger = _logging.adapt_logger(logger or _logging.NoOpLogger())

    def resolve_or_none(self, user):

        if user is None:
            return None

        if isinstance(user, User):
            hackle_user = HackleUser.of(user)
        elif isinstance(user, HackleUser):
            hackle_user = user
        else:
            return None

        if hackle_user.properties is None:
            user_properties = {}
        else:
            user_properties = hackle_user.properties

        if not validator.is_valid_properties(user_properties):
            self.logger.warning(
                'User properties is not valid. User properties must be not dic types and items types must be string_types, number, bool.'
                ': {}'.format(user_properties))
            user_properties = utils.filter_properties(user_properties)

        identifiers_builder = IdentifiersBuilder(self.logger)
        identifiers_builder.add_identifiers(hackle_user.identifiers)

        if hackle_user.id is not None:
            identifiers_builder.add(IdentifierType.ID, hackle_user.id)

        if hackle_user.user_id is not None:
            identifiers_builder.add(IdentifierType.USER, hackle_user.user_id)

        if hackle_user.device_id is not None:
            identifiers_builder.add(IdentifierType.DEVICE, hackle_user.device_id)

        identifiers = identifiers_builder.build()

        if len(identifiers) == 0:
            return None

        return InternalHackleUser(identifiers, user_properties)



import abc
import time

from hackle.user.identifier_type import IdentifierType

ABC = abc.ABCMeta('ABC', (object,), {'__slots__': ()})


class UserEvent(ABC):
    def __init__(self, user):
        self.timestamp = self._get_time()
        self.userId = user.identifiers.get(IdentifierType.ID)
        self.identifiers = user.identifiers
        self.userProperties = user.properties

    # noinspection PyMethodMayBeStatic
    def _get_time(self):
        return int(round(time.time() * 1000))

    @staticmethod
    def exposure(experiment, user, evaluation):
        return ExposureEvent(user, experiment, evaluation)

    @staticmethod
    def track(event_type, event, user):
        return TrackEvent(user, event_type, event)


class ExposureEvent(UserEvent):
    def __init__(self, user, experiment, evaluation):
        super(ExposureEvent, self).__init__(user)
        self.experimentId = experiment.id
        self.experimentKey = experiment.key
        self.experimentType = experiment.type
        self.experimentVersion = experiment.version
        self.variationId = evaluation.variation_id
        self.variationKey = evaluation.variation_key
        self.decisionReason = evaluation.reason


class TrackEvent(UserEvent):
    def __init__(self, user, event_type, event):
        super(TrackEvent, self).__init__(user)
        self.eventTypeId = event_type.id
        self.eventTypeKey = event_type.key
        self.value = event.value
        self.properties = event.properties



class Evaluation(object):
    def __init__(self, variation_id, variation_key, reason):
        self.variation_id = variation_id
        self.variation_key = variation_key
        self.reason = reason

    def __eq__(self, o):
        if isinstance(o, self.__class__):
            return self.__dict__ == o.__dict__
        else:
            return False

    def __str__(self):
        return 'Evaluation(variation_id={}, variation_key={}, reason={})'.format(self.variation_id,
                                                                                 self.variation_key,
                                                                                 self.reason)

    def __repr__(self):
        return self.__str__()

    @staticmethod
    def with_key(variation_key, reason):
        return Evaluation(variation_id=None, variation_key=variation_key, reason=reason)

    @staticmethod
    def with_variation(variation, reason):
        return Evaluation(variation_id=variation.id, variation_key=variation.key, reason=reason)

    @staticmethod
    def of(experiment, variation_key, reason):
        variation = experiment.get_variation_by_key_or_none(variation_key)
        if variation:
            return Evaluation.with_variation(variation, reason)
        else:
            return Evaluation.with_key(variation_key, reason)


class Evaluator(object):

    def __init__(self, evaluation_flow_factory):
        self.evaluation_flow_factory = evaluation_flow_factory

    def evaluate(self, workspace, experiment, user, default_variation_key):
        evaluation_flow = self.evaluation_flow_factory.get_evaluation_flow(experiment.type)
        return evaluation_flow.evaluate(workspace, experiment, user, default_variation_key)



class BaseEntity(object):
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False


class Experiment(BaseEntity):

    def __init__(self, id, key, type, identifier_type, status, version, variations, user_overrides, segment_overrides,
                 target_audiences, target_rules, default_rule, container_id, winner_variation_id):
        self.id = id
        self.key = key
        self.type = type
        self.identifier_type = identifier_type
        self.status = status
        self.version = version
        self.variations = variations
        self.user_overrides = user_overrides
        self.segment_overrides = segment_overrides
        self.target_audiences = target_audiences
        self.target_rules = target_rules
        self.default_rule = default_rule
        self.container_id = container_id
        self.winner_variation = self.get_variation_by_id_or_none(winner_variation_id)

    def get_variation_by_id_or_none(self, variation_id):
        for variation in self.variations:
            if variation.id == variation_id:
                return variation
        return None

    def get_variation_by_key_or_none(self, variation_key):
        for variation in self.variations:
            if variation.key == variation_key:
                return variation
        return None


class Variation(BaseEntity):
    def __init__(self, id, key, is_dropped):
        self.id = id
        self.key = key
        self.is_dropped = is_dropped


class Bucket(BaseEntity):
    def __init__(self, seed, slot_size, slots):
        self.seed = seed
        self.slot_size = slot_size
        self.slots = slots

    def get_slot_or_none(self, slot_number):
        for slot in self.slots:
            if slot.contains(slot_number):
                return slot
        return None


class Slot(BaseEntity):
    def __init__(self, start_inclusive, end_exclusive, variation_id):
        self.start_inclusive = start_inclusive
        self.end_exclusive = end_exclusive
        self.variation_id = variation_id

    def contains(self, slot_number):
        return (self.start_inclusive <= slot_number) and (slot_number < self.end_exclusive)


class EventType(BaseEntity):
    def __init__(self, id, key):
        self.id = id
        self.key = key


class Target(BaseEntity):
    def __init__(self, conditions):
        self.conditions = conditions


class TargetCondition(BaseEntity):
    def __init__(self, key, match):
        self.key = key
        self.match = match


class TargetKey(BaseEntity):
    def __init__(self, type, name):
        self.type = type
        self.name = name


class TargetMatch(BaseEntity):
    def __init__(self, type, operator, value_type, values):
        self.type = type
        self.operator = operator
        self.value_type = value_type
        self.values = values


class TargetAction(BaseEntity):
    def __init__(self, type, variation_id=None, bucket_id=None):
        self.type = type
        self.variation_id = variation_id
        self.bucket_id = bucket_id


class TargetRule(BaseEntity):
    def __init__(self, target, action):
        self.target = target
        self.action = action


class Segment(BaseEntity):
    def __init__(self, id, key, type, targets):
        self.id = id
        self.key = key
        self.type = type
        self.targets = targets


class TargetingType(object):

    def __init__(self, *supported_key_types):
        self.supported_key_types = supported_key_types

    def supports(self, target_key_type):
        return target_key_type in self.supported_key_types


class TargetingTypes(object):
    IDENTIFIER = TargetingType('SEGMENT')
    PROPERTY = TargetingType('SEGMENT', 'USER_PROPERTY', 'HACKLE_PROPERTY')
    SEGMENT = TargetingType('USER_ID', 'USER_PROPERTY', 'HACKLE_PROPERTY')


class Container(object):

    def __init__(self, id, bucket_id, groups):
        self.id = id
        self.bucket_id = bucket_id
        self.groups = groups

    def get_group_or_none(self, container_group_id):
        for group in self.groups:
            if group.id == container_group_id:
                return group
        return None


class ContainerGroup(object):
    def __init__(self, id, experiments):
        self.id = id
        self.experiments = experiments
